/* Copyright (C) 2025 maClara, LLC <info@maclara-llc.com>
   This file is part of the JWT C Library

   SPDX-License-Identifier:  MPL-2.0
   This Source Code Form is subject to the terms of the Mozilla Public
   License, v. 2.0. If a copy of the MPL was not distributed with this
   file, You can obtain one at http://mozilla.org/MPL/2.0/. */

#ifndef JWT_OPENSSL_H
#define JWT_OPENSSL_H

int openssl_process_eddsa(jwt_json_t *jwk, jwk_item_t *item);
int openssl_process_rsa(jwt_json_t *jwk, jwk_item_t *item);
int openssl_process_ec(jwt_json_t *jwk, jwk_item_t *item);
void openssl_process_item_free(jwk_item_t *item);

#endif /* JWT_OPENSSL_H */
